/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package, developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL) and headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#if ( (defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)) \
        && (!defined(USE_MRAIL))  && (!defined(CH_GEN2_MRAIL)))

#include "mpid.h"
#include "mpiimpl.h"
#include "mpimem.h"

#if defined(CH_GEN2)
#include "../../mpid/ch_gen2/collutils.h"
#else
#include "../../mpid/vapi/collutils.h"
#endif

static inline void allgather_all_send_complete(int, int);

int intra_rdma_Allgather(void *sendbuf,
		int sendcount,
		struct MPIR_DATATYPE *sendtype,
		void *recvbuf,
		int recvcount,
		struct MPIR_DATATYPE *recvtype,
		struct MPIR_COMMUNICATOR *comm, int one_size)
{
	int dest, dimensions, mask, subcube_mask, rank, size, i, j;
	int alternate, next_alternate, round;
	int size_one_send_buf, size_one_recv_buf, virtual_colbuf_offset;
    int *peer_colbuf_offset = NULL;
    int timer=0,max_spin_time=10000;
	struct Collbuf *collbuf;
	char *local_addr, *dest_addr, *recv_addr;
	MPI_Aint sendtype_extent, recvtype_extent;
	void *poll_addr;
	char *local_buf = NULL;
	unsigned long allgather_msg_size, len;
	static unsigned long prev_allgather_msg_size = 0;
	static unsigned long clear_allgather_msg_size = 0;
	struct Allgather *allgather;
	struct Allgather_Peer *peers;
	struct L_Collbuf *l_coll;
	struct R_Collbuf *r_coll;
    struct MPIR_COMMUNICATOR *actual_comm;
	dreg_entry *d_recv;

	/* Get size and switch to collective communicator */
	MPIR_Comm_size(comm, &size);
	MPIR_Comm_rank(comm, &rank);

    actual_comm = comm;

	comm = comm->comm_coll;

	if (comm->collbuf == NULL) {
		/* No proper initialization */
		MPID_Abort(comm, -1, "Illegal collective", NULL);
	}

	/* Get extent of send and recv types */
	MPI_Type_extent(sendtype->self, &sendtype_extent);
	MPI_Type_extent(recvtype->self, &recvtype_extent);

	size_one_send_buf = sendcount*sendtype_extent;
	size_one_recv_buf = recvcount*recvtype_extent;

	/* 0 byte Allgather ? */
	if (size_one_send_buf <= 0) {
		return MPI_SUCCESS;
	}

	collbuf = comm->collbuf;
    /* Adjusting the offset for the allgather buffer*/

	l_coll = (collbuf->l_coll) ;
	r_coll = collbuf->r_coll;
	allgather = collbuf->allgather;
	dimensions = allgather->dimensions;
	peers = allgather->peers;
    peer_colbuf_offset = (int*) calloc(sizeof(int), dimensions);

    if(peer_colbuf_offset == NULL) {
        fprintf(stderr, "[%d] Failed to allocate peer_colbuf_offset file: %s, line: %d\n",
                rank, __FILE__, __LINE__);

        return MPI_ERR_UNKNOWN;
    }


	/* To support back-to-back allgather collective operations,
	 * we use three back-to-back buffers; no need to wait for sends to
	 * complete except to ensure progress */
	alternate = allgather->alternate;
	next_alternate = (alternate+1)%3;
	allgather->alternate = next_alternate;

	allgather_msg_size = size_one_send_buf + ALLGATHER_HEADER_SIZE;

	/* 1. Clear next b2b buffer as required */
	if (clear_allgather_msg_size > ALLGATHER_THRESHOLD)	{
		/* just clear the flags used for completion RDMA writes and 
		 * the space used for address as we had used only zero-copy */
		memset(l_coll->buf + ALLGATHER_OFFSET + next_alternate*COLL_SINGLE_BUF_SIZE
				+ ALLGATHER_BUF_SIZE, 0, 
				COLL_FLAG_SPACE + COLL_ADDR_RECV_SPACE);
	}else{
		memset(l_coll->buf + ALLGATHER_OFFSET + next_alternate*COLL_SINGLE_BUF_SIZE,
			0, MIN((clear_allgather_msg_size+ALLGATHER_HEADER_SIZE)*size, 
				COLL_SINGLE_BUF_SIZE));

		if (clear_allgather_msg_size*size/2 > ALLGATHER_THRESHOLD)	{
			/* we used zero-copy also */
			memset(l_coll->buf + ALLGATHER_OFFSET + next_alternate*COLL_SINGLE_BUF_SIZE
				+ ALLGATHER_BUF_SIZE, 0, 
				COLL_FLAG_SPACE + COLL_ADDR_RECV_SPACE);
		}
	}
	local_buf = l_coll->buf + ALLGATHER_OFFSET + alternate*COLL_SINGLE_BUF_SIZE;

	/* 2. Register the recv buffer and copy the recv buffer address
	 * to the collective buffer to be sent out if required */
	if (size_one_recv_buf*size/2 > ALLGATHER_THRESHOLD)	{

		/* Register the user buffer */
		d_recv = dreg_register(recvbuf, size_one_recv_buf*comm->np,
                DREG_ACL_WRITE);

		if (NULL == d_recv) {
			fprintf(stderr, "[%d] Registration failed, file: %s, line: %d\n",
					rank, __FILE__, __LINE__);
			return MPI_ERR_UNKNOWN;
		}

		/* copy your user buffer address to collective buffer */
#if defined(CH_GEN2)
        copy_recv_addr_allgather(recvbuf, collbuf, d_recv->memhandle);
#else
		copy_recv_addr_allgather(recvbuf, collbuf, &(d_recv->memhandle));
#endif
	}

	/* 3. Send address of user buffer if required and that too to only
	 * necessary peers */
	virtual_colbuf_offset = 0;
	subcube_mask = (1 << dimensions) - 1;
    for(mask=0x1,round=0; mask<size; mask<<=1,round++){
        dest = rank^mask;
        len = mask*size_one_recv_buf;

        if (len > ALLGATHER_THRESHOLD)	{
            if (0 == virtual_colbuf_offset)	{
                virtual_colbuf_offset = (rank&subcube_mask)*allgather_msg_size;
                for (j=0x1, i=0; i<dimensions; j<<=1, i++)
                    peer_colbuf_offset[i] = 
                        ((rank^j)&subcube_mask)*allgather_msg_size;
            }

#if defined(CH_GEN2)
            coll_rdma_write(actual_comm->lrank_to_grank[dest],
                    allgather->buf_hndl_addr,
                    (allgather->buf_hndl_addr_len +
                     allgather->rdma_recv_addr_len + sizeof(char)),
                    l_coll->mr,
                    peers[round].dest_addr[alternate],
                    (r_coll[dest].rkey));
#else        
            coll_rdma_write(actual_comm->lrank_to_grank[dest],
                    allgather->buf_hndl_addr,
                    allgather->buf_hndl_addr_len +
                    allgather->rdma_recv_addr_len +
                    sizeof(char), 
                    &(l_coll->buf_hndl),
                    peers[round].dest_addr[alternate],
                    &(r_coll[dest].buf_hndl));
#endif
        }
        subcube_mask <<= 1;
    }

	local_buf = local_buf - virtual_colbuf_offset;

	/* 4. Copy over my message to my collective buffer and recv buffer */
	if (size_one_recv_buf <= ALLGATHER_THRESHOLD)	{
		copy_to_rdma_buf(local_buf + rank*allgather_msg_size, sendbuf, 
				size_one_send_buf, rank, rank);
	}
	memcpy(recvbuf + rank*size_one_recv_buf, sendbuf, size_one_recv_buf);


	/* 5. Do allgather using recursive doubling */
	subcube_mask = (1 << dimensions) - 1;

	for(mask=0x1, round=0; mask<size; mask<<=1, round++)	{
		dest = rank^mask;

		if (mask*size_one_send_buf <= ALLGATHER_THRESHOLD)	{

			/* copy-based approach */
			len = mask*allgather_msg_size;
			local_addr = local_buf + (rank & subcube_mask)*allgather_msg_size;
			dest_addr = r_coll[dest].buf + ALLGATHER_OFFSET + alternate*COLL_SINGLE_BUF_SIZE +
				(rank & subcube_mask)*allgather_msg_size - 
				peer_colbuf_offset[round];

#if defined(CH_GEN2)
            coll_rdma_write(actual_comm->lrank_to_grank[dest],
                    local_addr, len,
                    (l_coll->mr), dest_addr,
                    (r_coll[dest].rkey));
#else
        
			coll_rdma_write(actual_comm->lrank_to_grank[dest], 
                    local_addr, len, &(l_coll->buf_hndl), 
					dest_addr, &(r_coll[dest].buf_hndl));
#endif

			/* overlap progress with largest send. Largest send occurs in
			 * last iteration: of p/2 messages */
			if (mask == size/2)	MPID_DeviceCheck(MPID_NOTBLOCKING);

			/* Recursive Doubling is a combining algo. Data from all
			 * previous iterations is needed for a given iteration */
			recv_addr = local_buf + (dest & subcube_mask)*allgather_msg_size;
			poll_addr = recv_addr + len - sizeof(char);
            
            timer = 0;
			while(*((volatile char *)poll_addr) != ALLGATHER_TAIL_FLAG){
                if (++timer >= max_spin_time){
                    timer=0;
                    MPID_DeviceCheck(MPID_NOTBLOCKING);
                }
            }

			/* Copy over the data to receive buffer now itself */
			for(i=0; i<mask; i++)	{
				memcpy(recvbuf + ((dest&subcube_mask)+i)*size_one_recv_buf, 
						recv_addr + i*allgather_msg_size + sizeof(char),
						size_one_recv_buf);
			}
		}else{


			/* zero-copy approach */
			len = mask*size_one_recv_buf;

			/* wait for his address */
#if defined(CH_GEN2)
			poll_addr = (void *) (peers[round].recv_addr[alternate] + 
                    sizeof(uint32_t) + sizeof(aint_t));
#else
			poll_addr = (void *) (peers[round].recv_addr[alternate] + 
                    sizeof(VAPI_rkey_t) + sizeof(aint_t));
#endif
            

            timer = 0;
            while(*((volatile char *)poll_addr) != ALLGATHER_RDMA_FLAG){
                if (++timer >= max_spin_time){
                    timer=0;
                    MPID_DeviceCheck(MPID_NOTBLOCKING);
                }
            }

            /* Reset Poll flag */
            *(char *) poll_addr = 0;

			/* Got his address; send him the data */
			local_addr = recvbuf + (rank & subcube_mask)*size_one_recv_buf;

#if defined(CH_GEN2)
			peers[round].mr.rkey =
				*((uint32_t*) peers[round].recv_addr[alternate]);
			dest_addr = (char *)(*(aint_t *)(peers[round].recv_addr[alternate] 
						+ sizeof(uint32_t)));
#else
			peers[round].memhandle.rkey =
				*((VAPI_rkey_t *) peers[round].recv_addr[alternate]);
			dest_addr = (char *)(*(aint_t *)(peers[round].recv_addr[alternate] 
						+ sizeof(VAPI_rkey_t)));
#endif
			dest_addr = dest_addr + (rank & subcube_mask)*size_one_recv_buf;

			/* write the data */
#if defined(CH_GEN2)
            coll_rdma_write(actual_comm->lrank_to_grank[dest],
                    local_addr, len, (d_recv->memhandle), dest_addr,
                    (peers[round].mr.rkey));
#else
                    
			coll_rdma_write(actual_comm->lrank_to_grank[dest], 
                    local_addr, len, &(d_recv->memhandle), 
					dest_addr, &(peers[round].memhandle));
#endif

			/* Indicate completion thru b2b RDMA write */
#if defined(CH_GEN2)
            coll_rdma_write(actual_comm->lrank_to_grank[dest],
                    allgather->rdma_flag,
                    allgather->rdma_flag_size,
                    (l_coll->mr),
                    peers[round].r_done[alternate],
                    (r_coll[dest].rkey));
#else
			coll_rdma_write(actual_comm->lrank_to_grank[dest], 
                    allgather->rdma_flag, 
					allgather->rdma_flag_size, 
					&(l_coll->buf_hndl), 
					peers[round].r_done[alternate], 
					&(r_coll[dest].buf_hndl));
#endif

			if (mask == size/2)	MPID_DeviceCheck(MPID_NOTBLOCKING);

			poll_addr = peers[round].done[alternate];
            timer=0;
            while(*((volatile char *)poll_addr) != ALLGATHER_RDMA_FLAG){
                if (++timer >= max_spin_time){
                    timer=0;
                    MPID_DeviceCheck(MPID_NOTBLOCKING);
                }
            }
		}
		subcube_mask <<= 1;
	}

	clear_allgather_msg_size = prev_allgather_msg_size;
	prev_allgather_msg_size = size_one_send_buf;

	/* 6. Unregister the user buffer if you had RDMA written from/to it */
	if (size_one_recv_buf*size/2 > ALLGATHER_THRESHOLD)	{
		allgather_all_send_complete(rank, size);
		dreg_unregister(d_recv);
	}

    free(peer_colbuf_offset);

	return (MPI_SUCCESS);
}


static inline void allgather_all_send_complete(int rank, int size)
{
	unsigned mask;
	int not_complete;
	viadev_connection_t *c = NULL;

	do{
		not_complete = 0;
		for (mask=0x1; mask<size; mask<<=1)	{
			c = &viadev.connections[rank^mask];
			if(c->send_wqes_avail != viadev_sq_size) {
				not_complete = 1;
				break;
			}
		}
		if (not_complete)	MPID_DeviceCheck(MPID_NOTBLOCKING);
	}while (not_complete);
}

#endif /* RDMA_ALLGATHER */
