/* Copyright (c) 2002-2008, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package, developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL) and headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */

#if ((defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)) && (!defined(CH_GEN2_MRAIL)))

#include "mpid.h"
#include "mpiimpl.h"
#include "mpimem.h"

#ifdef USE_MRAIL
#include "../../mpid/vapi_multirail/collutils.h"
#elif defined(CH_GEN2)
#include "../../mpid/ch_gen2/collutils.h"
#else
#include "../../mpid/vapi/collutils.h"
#endif

int intra_rdma_Alltoall(void *sendbuf,
                        int sendcount,
                        struct MPIR_DATATYPE *sendtype,
                        void *recvbuf,
                        int recvcnt,
                        struct MPIR_DATATYPE *recvtype,
                        struct MPIR_COMMUNICATOR *comm, int one_size)
{
    int size, i, rank, size_one_send_buf, size_one_recv_buf;
    MPI_Aint sendtype_extent, recvtype_extent, sendbuf_extent;
    int sendtype_size;
    dreg_entry *d_send = NULL;
    dreg_entry *d_recv = NULL;
    int dest_rank, alternate, next_alternate;
    struct MPIR_COMMUNICATOR *actual_comm;

    struct R_Collbuf *r_coll;
    struct L_Collbuf *l_coll;
    struct Alltoall *atoa;
    struct DE_Remote_Node *node;

    void *poll_addr_ptr;
    void *dest_addr;
    void *local_addr;


    /* Get size and switch to collective communicator */
    MPIR_Comm_size(comm, &size);
    MPIR_Comm_rank(comm, &rank);

    actual_comm = comm;
    comm = comm->comm_coll;

    if (comm->collbuf == NULL) {
        /* No proper initialization */
        MPID_Abort(comm, -1, "Illegal collective", NULL);
    }

    /* Get extent of send and recv types */
    MPI_Type_extent(sendtype->self, &sendtype_extent);
    MPI_Type_extent(recvtype->self, &recvtype_extent);

    MPI_Type_size(sendtype->self, &sendtype_size);

    size_one_send_buf = sendcount * sendtype_size;
    size_one_recv_buf = recvcnt * recvtype_extent;

    /* 0 byte Alltoall ? */
    if (size_one_send_buf <= 0) {
        return MPI_SUCCESS;
    }

    r_coll = comm->collbuf->r_coll;
    l_coll = comm->collbuf->l_coll;
    atoa = comm->collbuf->atoa;
    node = atoa->de->node;
    alternate = atoa->de->alternate;

    /* Toggle */
    atoa->de->alternate = (alternate ^ 0x01);

    next_alternate = atoa->de->alternate;

    /* 1. Register the receive buffer */
    d_recv = dreg_register(recvbuf, size_one_recv_buf * comm->np, 
            DREG_ACL_WRITE);

    if (NULL == d_recv) {
        fprintf(stderr, "[%d] Registration failed, file : %s, line : %d\n",
                rank, __FILE__, __LINE__);
        return MPI_ERR_UNKNOWN;
    }

    /* 2 . Copy your recv addr & key into the pre-registered buffer */
#if defined(CH_GEN2)
    copy_recv_addr(recvbuf, comm->collbuf, d_recv->memhandle);
#else
    copy_recv_addr(recvbuf, comm->collbuf, &(d_recv->memhandle));
#endif

    /* 3. Send your recv addr to your peers */
    for (i = 0; i < size; i++) {
        dest_rank = (rank + i) % size;

        /* Don't send it to myself ! */
        if (dest_rank == rank)
            continue;
#if defined(CH_GEN2)
        coll_rdma_write(actual_comm->lrank_to_grank[dest_rank], 
                atoa->buf_hndl_addr,
                (atoa->buf_hndl_addr_len +
                 atoa->rdma_recv_addr_len + sizeof(char)), 
                l_coll->mr,
                node[dest_rank].dest_addr[alternate],
                r_coll[dest_rank].rkey);
#else
        coll_rdma_write(actual_comm->lrank_to_grank[dest_rank],
                atoa->buf_hndl_addr,
                (atoa->buf_hndl_addr_len +
                 atoa->rdma_recv_addr_len + sizeof(char)),
                &(l_coll->buf_hndl),
                node[dest_rank].dest_addr[alternate],
                &(r_coll[dest_rank].buf_hndl));

#endif
    }

    /* 4. Set the alternate buffers to be zero. */
    for (i = 0; i < size; i++) {
        if (i == rank)
            continue;

        memset(node[i].recv_addr[next_alternate], 0,
               atoa->buf_hndl_addr_len + atoa->rdma_recv_addr_len
               + sizeof(char));

        *((char *) node[i].done[next_alternate]) = 0;
    }

    /* 5. Copy over my own buffer from send to recv */
    memcpy((void *) ((char *) recvbuf + (size_one_recv_buf * rank)),
           (void *) ((char *) sendbuf + (size_one_send_buf * rank)),
           size_one_send_buf);

    /* 6. Register the send buffer */
    d_send = dreg_register(sendbuf, size_one_send_buf * comm->np,
            DREG_ACL_WRITE);

    if (NULL == d_send) {
        fprintf(stderr, "[%d] Registration failed, file : %s, line : %d\n",
                rank, __FILE__, __LINE__);
        return MPI_ERR_UNKNOWN;
    }

    /* 7. Wait for other peers's addresses */
    for (i = (size - 1); i >= 0; i--) {
        dest_rank = (rank + i) % size;

        if (dest_rank == rank)
            continue;

#if defined(CH_GEN2)
        poll_addr_ptr =
            (void *) ((char *) node[dest_rank].recv_addr[alternate] +
                      sizeof(uint32_t) + sizeof(aint_t));
#else
        poll_addr_ptr =
            (void *) ((char *) node[dest_rank].recv_addr[alternate] +
                      sizeof(VAPI_rkey_t) + sizeof(aint_t));
#endif
        while (*((volatile char *) poll_addr_ptr) == 0);


        /* Fill in the data structures */
#if defined(CH_GEN2)
        node[dest_rank].mr.rkey =
            *((uint32_t *) node[dest_rank].recv_addr[alternate]);

        dest_addr =
            (void *) ((char *) node[dest_rank].recv_addr[alternate] +
                      sizeof(uint32_t));
#else
        node[dest_rank].memhandle.rkey =
            *((VAPI_rkey_t *) node[dest_rank].recv_addr[alternate]);

        dest_addr =
            (void *) ((char *) node[dest_rank].recv_addr[alternate] +
                      sizeof(VAPI_rkey_t));
#endif

        dest_addr = (void *) (*(aint_t *) dest_addr);

        /* Got this guy's address, should send him
         * my data immediately */

        /* My index in his buffer should be given by
         * my rank */
        dest_addr = (void *) ((char *) dest_addr +
                              rank * recvcnt * recvtype_extent);

        /* The local index should be given by
         * his rank */
        local_addr = (void *) ((char *) sendbuf +
                               dest_rank * sendcount * sendtype_extent);

        /* RDMA write */
#if defined(CH_GEN2)
        coll_rdma_write(actual_comm->lrank_to_grank[dest_rank],
                local_addr,
                size_one_send_buf,
                d_send->memhandle, dest_addr,
                node[dest_rank].mr.rkey);
#else
        coll_rdma_write(actual_comm->lrank_to_grank[dest_rank],
                local_addr,
                size_one_send_buf,
                &(d_send->memhandle), dest_addr,
                &(node[dest_rank].memhandle));
#endif
        /* RDMA write to tell him that I
         * am done with him */
#if defined(CH_GEN2)
        coll_rdma_write(actual_comm->lrank_to_grank[dest_rank], 
                atoa->rdma_flag,
                atoa->rdma_flag_size,
                l_coll->mr,
                node[dest_rank].r_done[alternate],
                r_coll[dest_rank].rkey);
#else
        coll_rdma_write(actual_comm->lrank_to_grank[dest_rank], 
                atoa->rdma_flag,
                atoa->rdma_flag_size,
                &(l_coll->buf_hndl),
                node[dest_rank].r_done[alternate],
                &(r_coll[dest_rank].buf_hndl));
#endif
    }


    /* Wait till everyone is done with me */
    for (i = 0; i < size; i++) {
        dest_rank = (rank + i) % size;

        if (dest_rank == rank)
            continue;

        poll_addr_ptr = (void *) node[dest_rank].done[alternate];

        while (*((volatile char *) poll_addr_ptr) == 0);

        /* Check to see if my sends to this process are complete */
        coll_all_send_complete(actual_comm->lrank_to_grank[dest_rank]);
    }

    /* Deregister the buffers */
    dreg_unregister(d_recv);
    dreg_unregister(d_send);

    return (MPI_SUCCESS);
}

#endif                          /* VIADEV_RPUT_SUPPORT */
